
#include <iostream>
#include <cuda.h>
#include "image.h"

// Helpers for checking errors from CUDA calls 
#define checkCudaErrors(val) check( (val), #val, __FILE__, __LINE__)

template<typename T>
void check(T err, const char* const func, const char* const file, const int line) {
   if (err != cudaSuccess) {
      std::cerr << "CUDA error at: " << file << ":" << line << std::endl;
      std::cerr << cudaGetErrorString(err) << " " << func << std::endl;
      exit(1);
   }
}


__global__ void AssignKernel(int* input, int* output, unsigned int n)
{
   unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

   if(i < n)
    output[i] = input[i];
}


int main(int argc, char **argv)
{
   if(argc != 2)
   {
      std::cerr << "Error: program <input-image-file>\n";
      return -1;
   }
   std::string infile = argv[1]; // "<image-file>.png"; // filename of image...
   
   #ifdef FREEIMAGE_LIB 
   FreeImage_Initialise();
   #endif

   FreeImage_SetOutputMessage(FreeImageErrorHandler);
   FIBITMAP *dib = NULL;

   dib = GenericLoader(infile.c_str(), 0);
   dib = FreeImage_ConvertToGreyscale(dib);
   
   unsigned width = FreeImage_GetWidth(dib);
   unsigned height = FreeImage_GetHeight(dib);

   unsigned pitch = FreeImage_GetPitch(dib);

   BYTE *bits = (BYTE*)FreeImage_GetBits(dib);
   
   int *input, *output;
   
   // code to allocate width*height memory for arrays...
   input = new int[width*height];
   output = new int[width*height];

   // following loop do conversion to int input array
   for(int y = 0; y < height; y++)
   {
      for(int x = 0; x < width; x++)
      {
         input[y*width+x] = bits[x];
      }
      bits += pitch;
   }
   
   int *d_input;
   int *d_output;
   size_t size = width*height*sizeof(int);
   checkCudaErrors(cudaMalloc(&d_input, size));
   checkCudaErrors(cudaMemcpy(d_input, input, size, cudaMemcpyHostToDevice));
   checkCudaErrors(cudaMalloc(&d_output, size));

   int n = width*height;
   int numThreads = (n<512)? n: 512;
   int numBlocks = (n + numThreads -1 ) / numThreads;
   AssignKernel<<<numBlocks, numThreads>>>(d_input, d_output, n)
   
   checkCudaErrors(cudaMemcpy(output, d_output, size, cudaMemcpyDeviceToHost));
   checkCudaErrors(cudaFree(d_input));
   checkCudaErrors(cudaFree(d_output));
   
   // Code to apply MapOverlap filter row and column wise to input array.
   
   // Code to convert it back to
   bits = (BYTE*)FreeImage_GetBits(dib);
   for(int y = 0; y < height; y++)
   {
      for(int x = 0; x < width; x++)
      {
         bits[x] = output[y*width+x];
      }
      bits += pitch;
   }
   
   // write to file back
   std::string outfile = "out_" + infile; // output filename...
   bool bSuccess = GenericWriter(dib, outfile.c_str() , PNG_DEFAULT);

   FreeImage_Unload(dib);

   #ifdef FREEIMAGE_LIB
   FreeImage_DeInitialise();
   #endif
   
   // code to free memory for arrays...
   delete [] input;
   delete [] output;

   return 0;
}